#!/usr/bin/env python3
#coding:utf-8
# Author:  Wolfram Eifler --<notiznetz@weifler.in-berlin.de>
# Purpose: graph-representation of a noteswebs data with incore-objects
# Created: 01/27/2017
# Copyright 2017, 2021 Wolfram Eifler
# Usage-Permission: GNU General Public License version 3 as published by the
# Free Software Foundation. The full text of the license is availabe
# in this distribution in the file licenses/gpl-v3.txt or
# via internet at http://www.gnu.org/licenses/gpl-3.0.html
"""
graph-representation of a noteswebs data with incore-objects
"""

import os, tempfile, unittest
import notiznetz
import notiznetz.store
from notiznetz.util import *
from notiznetz.nzz2svg import Drawing as svgDrawing, readnotes as svg_readnotes

class Note:
    """a Note in the Graph, represented via object-references
    (in contrast: in NoteData it is via ID-values"""
    def __init__(self, ident, parentid=0, title=None, text=None):
        if type(ident) == type(1): # ident needs to be integer
            self.ident = ident
            self.parentid = parentid    # for containment-relation
            if title != None:
                self.title = title
            else:
                self.title = "new Note_%i" % self.ident
            if text != None:
                self.text = text
            else:
                self.text = ""
            self.setHeadlineColour(0)
            self.setGeometry(-50,-25,100,50)    # center of screen
            self.setEditWindow([False, 100,100, 500,250, 0])
        else:
            print ("Error, Note without ident is not possible")

    def updateContainer(self, notesweb):
        """updates container-property (i am contained by) by querying
        notesweb for Note with ident of self.parentid"""
        assert(isinstance(notesweb, NotesWeb))
        self.container = notesweb.getNoteByID(self.parentid)

    def setHeadlineColour(self, colourindex):
        """sets the colour of the notes headline and headlinetext via index into
        MainView-global colour-table"""
        self.headlinecolour = colourindex

    def getGeometry(self):
        """returns (x, y, width, height) of this note"""
        return(self.x, self.y, self.width, self.height)

    def setGeometry(self, x, y, width=None, height=None, z=97):
        """ommitting  width/height will result in "setPos", current
        width and height will be kept. z determines stacking-order for
        drawing on canvas etc."""
        self.x = x
        self.y = y
        self.z = z
        if width != None: self.width = width
        if height != None: self.height = height

    def setEditWindow(self, editwindow_data):
        """editwindow_data is of form
        [bool:visible, x,y, width,height, scrollpos]"""
        self.editwin = editwindow_data

    def __str__(self):
        result = "Graph.Note: %i, '%s'\n" % (self.ident, self.title)
        result += "  %i, %i, %i, %i" % (self.x, self.y, self.width, self.height)
        return(result)

class Link:
    """a Note in the Graph, represented via object-knows-otherobject
    (in contrast: in LinkData it is via ID-values"""
    def __init__(self, ident, srcnote, dstnote=None, title=None):
        assert(isinstance(srcnote, Note))
        if dstnote != None: assert(isinstance(dstnote, Note))
        self.ident = ident
        self.src = srcnote
        self.dst = dstnote
        if title != None:
            self.title = title
        else:
            self.title = ""

    def connect(self, srcnote, dstnote):
        assert(isinstance(srcnote, Note))
        assert(isinstance(dstnote, Note))
        self.src = srcnote
        self.dst = dstnote

    def __str__(self):
        result = "graph.Link %i: \n" % self.ident
        result += "src:" + str(self.src)
        result += "dst:" + str(self.dst)
        return(result)

class NotesWeb:
    def __init__(self):
        self.notescounter = IdentCounter()
        self.linkcounter = IdentCounter()
        self.notes = {}
        self.links = {}
        self.labelcolours = LabelColours()

    def getAllNotes(self):
        return(self.notes.values())

    def getAllLinks(self):
        return(self.links.values())

    def getNoteByID(self, note_id):
        if note_id == 0:
            return(None)
        return(self.notes[note_id])

    def getLinkByID(self, link_id):
        return(self.links[link_id])

    def getLinksOfNote(self, note):
        """returns a list of all links, connected to a note
        (incoming and outgoing)"""
        assert(isinstance(note, Note))
        result = []
        for (ident, l) in self.links.items():
            assert(isinstance(l, Link))
            if (l.src == note) or (l.dst == note):
                result.append(l)
        return(result)

    def getLinksToNote(self, note):
        """returns a list of links pointing to a note"""
        result = []
        for l in self.getLinksOfNote(note):
            if l.dst == note:
                result.append(l)
        return(result)

    def getLinksFromNote(self, note):
        """returns a list of links from a note to another note"""
        result = []
        for l in self.getLinksOfNote(note):
            if l.src == note:
                result.append(l)
        return(result)

    def getContainedNotes(self, note):
        """finds all notes contained in note and returns them as list-of-Note.
        (containment is detected by Note.parentid)"""
        assert(isinstance(note, Note))
        result = []
        for n in self.getAllNotes():
            assert(isinstance(n, Note))
            if n.parentid == note.ident:
                result.append(n)
        return(result)

    def addNote(self, note):
        self.notes[note.ident] = note

    def addLink(self, link):
        self.links[link.ident] = link

    def newNote(self, title=None, text=None):
        n = Note(self.notescounter.getID(), parentid=0,
                 title=title, text=text)
        n.updateContainer(self)
        self.addNote(n)
        return(n)

    def newLink(self, srcnote, dstnote=None):
        assert(isinstance(srcnote, Note))
        if dstnote != None: assert(isinstance(dstnote, Note))
        l = Link(self.linkcounter.getID(), srcnote, dstnote)
        self.addLink(l)
        return(l)

    def deleteLink(self, link):
        assert(isinstance(link, Link))
        # remove from graph-global link-dict
        self.links.pop(link.ident, None)

    def deleteNoteAttachedLinks(self, note):
        assert(isinstance(note, Note))
        for l in self.getLinksOfNote(note):
            self.deleteLink(l)

    def deleteNote(self, note):
        assert(isinstance(note, Note))
        self.notes.pop(note.ident, None)

    def moveNoteInto(self, movingnote, containernote):
        """moves movingnote to be contained by containernote"""
        assert(isinstance(movingnote, Note))
        assert(isinstance(containernote, Note))
        movingnote.parentid = containernote.ident
        movingnote.updateContainer(self)

    def load(self, filename):
        """loads the contents of filename to own data via the
        store-module"""
        self.__init__() # re-initialize own vars.
        s = notiznetz.store.Store() # need the long name to avoid circular imports
        s.load(filename)
        self.notescounter._setCurrent(s.maxNoteID)
        self.linkcounter._setCurrent(s.maxLinkID)
        self.notes = s.notes
        self.links = s.links
        self.labelcolours = s.labelcolours
        for n in self.notes.values():
            assert(isinstance(n, Note))
            n.updateContainer(self)
        print ("loaded '%s, %i notes.'" % (filename,len(self.notes.keys())))
        print ("%i notes, %i links loaded from '%s'." % (
                len(self.notes.keys()), len(self.links.keys()), filename))
        return(s.mainWinData)

    def toStore(self, mainwindata):
        print ('----------- NotesWeb.toStore')
        s = notiznetz.store.Store() # need the long name to avoid circular imports
        s.labelcolours = self.labelcolours
        s.putNoteID(self.notescounter)
        s.putLinkID(self.linkcounter)
        s.putMainwinData(mainwindata)
        for n in self.getAllNotes():
            s.addNote(n)
        for l in self.getAllLinks():
            # only save link if both ends exist
            if l.src and l.dst and (l.src.ident in self.notes and
                l.dst.ident in self.notes):
                s.addLink(l)
            else:
                print ("Warning: toStore: found open link, ignored, not saved.")
                print (l.src)
                print (l.dst)
        return(s)

    def save(self, filename, mainwindata):
        """saves own data to filename using the store-module, also adds
        geometry-data for main-window to store.
        mainwindata is [x, y, width, height, scrollx, scrolly, zoomfactor]"""
        self.toStore(mainwindata).save(filename)

    def export_svg(self, filename, mainwindata):
        """save data as nzz-file and convert it"""
	# d = notiznetz.nzz2svg.Drawing(readnotes(sys.argv[1]), sys.argv[2])
        #
        tf = tempfile.NamedTemporaryFile()
        self.toStore(mainwindata).save(tf.name)
        tf.flush()
        # unlink tmpfile is done implicitely by close of tmpfile in nzz2svg
	# TODO: use a python-call in nzz2svg like
        d = svgDrawing(svg_readnotes(tf.name), filename)
        d.save()
        #returncode = os.system("nzz2svg %s %s" % (tf.name, filename))
        #if returncode != 0:
            #print ("ERROR ERROR ERROR something went wrong when calling the svg-converter 'nzz2svg' ")

    def __str__(self):
        result = "NotesWeb: %i Notes, %i Links\n" % (
            len(self.notes), len(self.links))
        result += "    nc=%i, lc=%i" % (
            self.notescounter._getCurrent(), self.linkcounter._getCurrent())
        return(result)

if __name__ == '__main__':
    unittest.main()
