#!/usr/bin/env python3
#coding:utf-8
# Author:  Wolfram Eifler --<notiznetz@weifler.in-berlin.de>
# Purpose: Main Window of the program
# Created: 01/22/2017
# Copyright 2017, Wolfram Eifler
# Usage-Permission: GNU General Public License version 3 as published by the
# Free Software Foundation. The full text of the license is availabe
# in this distribution in the file licenses/gpl-v3.txt or
# via internet at http://www.gnu.org/licenses/gpl-3.0.html
"""
Main Window of the program
"""

import sys, os, time
from PyQt5 import Qt, QtGui, QtWidgets, QtCore
from shutil import copyfile
import notiznetz
from notiznetz import Ui_MainWindow
from notiznetz.edittext import EditView
from notiznetz.widgets import NoteWidget, LinkWidget
from notiznetz.scene import NoteScene
from notiznetz.util import *

class MainView(QtWidgets.QMainWindow, Ui_MainWindow):
    """in MVC this Class is the View-Component"""
    def __init__(self, parent=None):
        super(MainView, self).__init__(parent)
        self.setupUi(self)
        assert(isinstance(self.graphicsView, Qt.QGraphicsView))
        self.current_filename = None
        self.labelcolours = LabelColours()
        self.secondary_views = []
        # make a scene and display it on graphicsview
        self.scene = NoteScene(self)
        self.toplevel_scene = self.scene
        #self.openEditViews = {}
        self.graphicsView.setScene(self.scene)
        self.colouractions = self._createColourActions()
        #
        # wire menu-actions
        self.actionCloseFile.triggered.connect(self.closeFile)
        self.action_NewFile.triggered.connect(self.newFile)
        self.action_Quit.triggered.connect(self.closeProgram)
        self.action_Open.triggered.connect(self.loadFileDialog)
        self.action_Save.triggered.connect(self.save)
        self.actionSave_As.triggered.connect(self.saveAsDialog)
        self.actionExport_SVG.triggered.connect(self.exportSVGDialog)
        self.actionAbout.triggered.connect(self.aboutDialog)
        self.actionUserdoc.triggered.connect(self.userdocDialog)
        self.actionNew_MapView.triggered.connect(self.newMapView)
        self.actionNewNote.triggered.connect(self.newNote)
        self.actionNotesText.triggered.connect(self.editSelectedNotes)
        #
        # make menu for note-headlines (not strictly OO here but more effective)
        self.menuContextNoteHeadline = Qt.QMenu(self)
        self.menuContextNoteHeadline.setObjectName("menuContextNoteHeadline")
        # actions for raise and lower
        self.actionRaiseNote = QtWidgets.QAction(self)
        self.actionRaiseNote.setObjectName("actionRaiseNote")
        self.actionRaiseNote.setText(QtCore.QCoreApplication.translate(
            "MainWindow", "Raise Note"))
        self.actionRaiseNote.triggered.connect(self.raiseNote)
        self.actionLowerNote = QtWidgets.QAction(self)
        self.actionLowerNote.setObjectName("actionLowerNote")
        self.actionLowerNote.setText(QtCore.QCoreApplication.translate(
                "MainWindow", "Lower Note"))
        self.actionLowerNote.triggered.connect(self.lowerNote)
        #
        self.menuContextNoteHeadline.addAction(self.actionRaiseNote)
        self.menuContextNoteHeadline.addAction(self.actionLowerNote)
        self.menuContextNoteHeadline.addAction(self.menuHeadcolour.menuAction())
        #
        self.addCrosshair(0,0,10)        # for orientation
        self.graphicsView.setDragMode(self.graphicsView.RubberBandDrag)

    def setHeadlineColour(self, colourindex):
        """sets the headlinecolour of the currently selected Notes to the
        colour indicated by colourindex (which is an integer pointing into
        labelcolours-table)."""
        self.scene.setHeadlineColour(colourindex)

    # ugly but working, better:
    # either setHeadlineColour can find out which action triggered it
    # or construct this functions via functional-style and currying
    #
    def _hCol0(self): self.setHeadlineColour(0)
    def _hCol1(self): self.setHeadlineColour(1)
    def _hCol2(self): self.setHeadlineColour(2)
    def _hCol3(self): self.setHeadlineColour(3)
    def _hCol4(self): self.setHeadlineColour(4)
    def _hCol5(self): self.setHeadlineColour(5)
    def _hCol6(self): self.setHeadlineColour(6)
    def _hCol7(self): self.setHeadlineColour(7)
    def _hCol8(self): self.setHeadlineColour(8)
    def _hCol9(self): self.setHeadlineColour(9)
    def _hCol10(self): self.setHeadlineColour(10)

    def _createColourActions(self):
        # create colour-actions
        result = []
        action_functions = [
            self._hCol0,
            self._hCol1,
            self._hCol2,
            self._hCol3,
            self._hCol4,
            self._hCol5,
            self._hCol6,
            self._hCol7,
            self._hCol8,
            self._hCol9,
            self._hCol10,
        ]
        i = 0
        for (bgcol, txtcol, txt) in self.labelcolours:
            act = QtWidgets.QAction(self)
            act.setObjectName("%s_%i" % (txt,i))
            self.menuHeadcolour.addAction(act)
            act.setText(QtCore.QCoreApplication.translate(
                "MainWindow", txt))
            act.triggered.connect(action_functions[i])
            result.append(act)
            i += 1
        return(result)

    def raiseNote(self):
        print ("raise-note called")
        self.scene.adjustNoteStacking(10)

    def lowerNote(self):
        print ("lower-note called")
        self.scene.adjustNoteStacking(-10)

    def visibleCenter(self):
        """get the center of visible"""
        gv = self.graphicsView
        c = gv.mapToScene(gv.viewport().rect().center())
        return(c)

    def addCrosshair(self, x, y, size=15):
        """draw crosshair-cursor at x,y, size is the length of the lines"""
        l1 = QtWidgets.QGraphicsLineItem(-size,y,size,y)
        l1.setPen(pen_b0)
        self.scene.addItem(l1)
        l2 = QtWidgets.QGraphicsLineItem(x,-size,x,size)
        l2.setPen(pen_b0)
        self.scene.addItem(l2)

    def getScreenPosition(self):
        t = self.graphicsView.transform()
        scalex = t.m11()    # strange qt-naming for scaling-factors
        scaley = t.m22()
        scrollx = self.graphicsView.verticalScrollBar().value()
        scrolly = self.graphicsView.horizontalScrollBar().value()
        #print (self.geometry())
        return(
            [self.x(), self.y(), self.width(), self.height(),
             scrollx, scrolly, scalex]
        )
        #[0, 348, 708, 394, -163, -157, 1.0]
        # [x,y,height,width, scrollx, scrolly, zoom]
        # assuming scalex is representative for zooming-factor
        # this is for use to setGeometry when starting next time with a file

    def setScreenPosition(self, geometry):
        """set window-position on screen, geometry-format is
        [x,y,height,width, scrollx, scrolly, zoomfactor]
        """
        (x,y,height,width, scrollx, scrolly, zoomfactor) = geometry
        self.setGeometry(x,y,height,width)
        self.graphicsView.scale(zoomfactor, zoomfactor)
        self.graphicsView.horizontalScrollBar().setValue(scrollx)
        self.graphicsView.verticalScrollBar().setValue(scrolly)

    def zoomIn(self):
        self.graphicsView.scale(1.05, 1.05)

    def zoomSetStd(self):
        print (self.getScreenPosition())
        #[0, 348, 708, 394, -163, -157, 1.0]
        self.graphicsView.resetTransform()
        self.setGeometry(0,0,708,500)

    def zoomOut(self):
        self.graphicsView.scale(1/1.05, 1/1.05)


    def openTextEdit(self, notewidget):
        """opens an Edit-Windows on a particular Note.
        To avoid multiple independent edits of the same note it just brings
        up the current Edit-Window if already one exists"""
        assert(isinstance(notewidget, NoteWidget))
        if notewidget.editwin:
            ev = notewidget.editwin
            ev.closeWindow()
        else:
            ev = EditView(notewidget, self)
        #if note.data.ident in self.openEditViews:
            #ev = self.openEditViews[note.data.ident]
            #ev.closeWindow()
        #else:
            #ev = EditView(note, self)
            #self.openEditViews[note.data.ident] = ev
        notewidget.setEditWindow(ev)
        (visibla, x,y,height,width, scrollpos) = notewidget.data.editwin
        ev.setGeometry(x, y, height, width)
        ev.editor.verticalScrollBar().setValue(scrollpos)
        ev.show()

    def editSelectedNotes(self):
        for note in self.scene.selectedNotes():
            self.openTextEdit(note)

    def newNote(self):
        self.scene.newNote()

    def save(self):
        if self.current_filename:
            self.toplevel_scene.save(self.current_filename)
        else:
            self.saveAsDialog()
        # remove data-changed-mark
        wt = self.windowTitle()
        if self.dataChanged():
            self.setWindowTitle(wt[:-2])

    def load(self, filename=None):
        if filename:
            self.setCurrentFilename(filename)
        self.scene = notiznetz.scene.NoteScene(self)
        self.toplevel_scene = self.scene
        self.graphicsView.setScene(self.scene)
        geom = self.scene.load(self.current_filename)
        self.setScreenPosition(geom)
        self.labelcolours = self.scene.data.labelcolours

    def dataChanged(self):
        return(self.windowTitle().endswith(' *'))

    def signalDataChange(self):
        wt = self.windowTitle()
        if not self.dataChanged():
            self.setWindowTitle(wt + ' *')

    def setCurrentFilename(self, filename):
        self.current_filename = filename
        self.setWindowTitle("NZZ: %s" % filename)

    def closeFile(self):
        if self.dataChanged():
            # ask if user wants to save his data
            self.dataChangedDialog()
        self.scene = notiznetz.scene.NoteScene(self)
        self.toplevel_scene = self.scene
        self.addCrosshair(0,0,10)
        self.graphicsView.setScene(self.scene)
        #self.scene.clear()  # also clears graph-data-structures
        # IMPORTANT!! would overwrite the last file when next ctrl-s happens else
        self.setCurrentFilename(None)

    def newFile(self):
        self.closeFile()
        self.save()

    def closeProgram(self):
        self.closeFile()
        self.close()

    def closeEvent(self, event):
        """intercept window-manager-close-event to check if data changed"""
        self.closeProgram()

    def loadFileDialog(self):
        if self.dataChanged():
            print ("data changed, ask for save before loading other file")
            self.dataChangedDialog()
        (filename, bla) = QtWidgets.QFileDialog.getOpenFileName(
            self, "Select Notes-File to load",
            filter='Notizzettel (*.nzz);;All Files (*)')
        if filename:
            self.load(filename)

    def saveAsDialog(self):
        (filename, bla) = QtWidgets.QFileDialog.getSaveFileName(
            self,"Select or type name of notizzettel-file to save",
            filter='Notizzettel (*.nzz);;All Files (*)')
        if filename:
            if not filename.endswith('.nzz'):
                filename = filename + '.nzz'
            self.setCurrentFilename(filename)
            self.save()

    def exportSVGDialog(self):
        (filename, bla) = QtWidgets.QFileDialog.getSaveFileName(
            self,"Select or type name of SVG-file to save",
            filter='SVG Vectorgraphics (*.svg);;All Files (*)')
        if filename:
            if not filename.endswith('.svg'):
                filename = filename + '.svg'
            self.toplevel_scene.export_svg(filename)

    def dataChangedDialog(self):
        mb = QtWidgets.QMessageBox
        reply = mb.question(self, 'Data Changed!',
                            "Current Filename is '%s'\n" % self.current_filename +
                            "Do you want to save it before closing?",
                            mb.Yes | mb.No, mb.Yes)
        if reply == mb.Yes:
            self.save()

    def aboutDialog(self):
        msg = "program-version:   %s\n" % notiznetz.__version__
        msg += "fileformat-version: %s\n" % notiznetz.fileformat_version
        msg += "License:\n"
        msg += "http://www.gnu.org/licenses/gpl-3.0.html\n"
        msg += "a local copy is in <installdir>/licenses/gpl-v3.txt"
        mb = QtWidgets.QMessageBox(QtWidgets.QMessageBox.Information,
                        "About notiznetz","notiznetz", QtWidgets.QMessageBox.Ok, self)
        mb.setInformativeText(msg)
        pathpre = os.path.split(__file__)[0]
        try:
            input = open(os.path.join(pathpre, 'licenses', 'gpl-v3.txt'), 'r', encoding="utf-8")
            gpltext = input.read()
            input.close()
        except:
            gpltext = "Problem reading license-file for gpl-v3.txt"
            sys.stdout.write("\nERROR: %s\n" % gpltext)
            sys.stdout.flush()
        mb.setDetailedText(gpltext)
        print ("position: %s" % __file__)
        mb.show()

    def userdocDialog(self):
        installdir = os.path.split(__file__)[0]
        docdir = os.path.join(installdir, 'doc')
        helpsrc = os.path.join(docdir, 'user.nzz')
        helpdst = '/tmp/nzz-userdoc.nzz'
        try:
            copyfile(helpsrc, helpdst)
            ok = True
        except:
            sys.stdout.write("\nERROR: Can not open Helpfile '%s'" % helpsrc)
            sys.stdout.flush()
            ok = False
        if ok:
            helpviewer = MainView()
            helpviewer.load(helpdst)
            helpviewer.show()

    def newMapView(self):
        newview = MainView()
        newview.setCurrentFilename(self.current_filename)
        # make a scene and display it on graphicsview
        newview.graphicsView.setScene(self.scene)
        self.secondary_views = [newview]
        newview.show()

if __name__ == '__main__':
    pass
