#!/usr/bin/env python3
#coding:utf-8
# Author:  Wolfram Eifler --<notiznetz@weifler.in-berlin.de>
# Purpose: convert an nzz-file to svg Vectorgraphics (only image, no text)
# Created: 01/16/2020
# Copyright 2020, 2021 Wolfram Eifler
# Usage-Permission: GNU General Public License version 3 as published by the
# Free Software Foundation. The full text of the license is availabe
# in this distribution in the file licenses/gpl-v3.txt or
# via internet at http://www.gnu.org/licenses/gpl-3.0.html
"""
convert an nzz-file to svg Vectorgraphics only image of linked cards and their
titles appear, not the texts from the notes-content)
"""


import sys, math, svgwrite
from svgwrite import cm, mm

import jsonpickle


# set prettyprinting for json-encoders so sourcecode-controlsystems can
# do sensible diffs on data-files
jsonpickle.set_encoder_options('simplejson', sort_keys=True, indent=2)
jsonpickle.set_encoder_options('json', sort_keys=True, indent=2)
jsonpickle.set_encoder_options('demjson', compactly=False)

class Point:
    def __init__(self, x, y):
        self.x = x
        self.y = y

    def distance(self, other):
        """Abstand zwischen zwei Punkten"""
        assert(isinstance(other, Point))
        return(math.sqrt( (other.x - self.x)**2 + (other.y - self.y)**2 ))

    def __str__(self):
        return("Point: x=%f y=%f" %  (self.x, self.y))

class Line:
    def __init__(self, x1, y1, x2, y2):
        self.start = Point(x1,y1)
        self.end = Point(x2, y2)
        if x1 < x2:
            self.minx = x1
            self.maxx = x2
        else:
            self.minx = x2
            self.maxx = x1
        if  y1 < y2:
            self.miny = y1
            self.maxy = y2
        else:
            self.miny = y2
            self.maxy = y1

    def length(self):
        return(self.start.distance(self.end))

    def midpoint(self):
        mx = (self.start.x + self.end.x)/2
        my = (self.start.y + self.end.y)/2
        return(Point(mx, my))

    def intersect(self, other):
        # ---------------------------------------------
        # Literatur für Intersection-Test (Stichwort Orientierung)
        # ---------------------------------------------
        # http://www.jeffreythompson.org/collision-detection/line-line.php
        # https://www.geeksforgeeks.org/check-if-two-given-line-segments-intersect/
        # https://algorithmtutor.com/Computational-Geometry/Check-if-two-line-segment-intersect/
        # ---------------------------------------------
        assert(isinstance(other, Line))
        x1 = self.start.x;    y1 = self.start.y
        x2 = self.end.x;     y2 = self.end.y
        x3 = other.start.x; y3 = other.start.y
        x4 = other.end.x;  y4 = other.end.y
        try:
            uA = ((x4-x3)*(y1-y3) - (y4-y3)*(x1-x3)) / ((y4-y3)*(x2-x1) - (x4-x3)*(y2-y1))
            uB = ((x2-x1)*(y1-y3) - (y2-y1)*(x1-x3)) / ((y4-y3)*(x2-x1) - (x4-x3)*(y2-y1))
        except:
            return(None) # division by zero -> no intersection

        # uA und  uB between 0 and 1 -> linesegments intersect
        if ((uA >= 0) and (uA <= 1) and (uB >= 0) and (uB <= 1)):
            # intersection-point
            intersect_x = x1 + (uA * (x2-x1))
            intersect_y = y1 + (uA * (y2-y1))
            return(Point(intersect_x, intersect_y))
        return(None)

    def __str__(self):
        result = "Line: x1=%f y1=%f | x2=%f y2=%f\n" % (
            self.start.x, self.start.y, self.end.x, self.end.y)
        result += "\tminx=%f maxx=%f miny=%f maxy=%f" % (
            self.minx, self.maxx, self.miny, self.maxy)
        return(result)

class Note:
    def __init__(self, datadict, headcolours):
        self.id = datadict['id'] # ggf. verwerten
        self.headcolour = headcolours[datadict['headcolour']]
        self.height = datadict['height']
        self.width = datadict['width']
        self.x = datadict['x']
        self.y = datadict['y']
        self.z = datadict['z']
        self.title = datadict['title']
    def addto(self, drawing, parentgroup):
        """add self to an svg-drawing"""
        headhight = (self.height/5)+4
        g = drawing.g(id=self.id)
        # Außenrahmen
        r = drawing.rect(
                insert=(self.x*mm, self.y*mm),
                size=(self.width*mm, self.height*mm),
                class_='notebase')
        g.add(r)
        # Hintergrundfarbe für Titelzeile [bgcol, txtcol, colourname]
        headbgcol = "#%s" % self.headcolour[0]
        bg = drawing.rect(
                     insert=(self.x*mm, self.y*mm),
                     size=(self.width*mm, headhight*mm),
                     fill=headbgcol, class_='headbase')
        g.add(bg)
        # Titeltext
        # Fontfarbe [bgcol, txtcol, colourname]
        txtcol = "#%s" % self.headcolour[1]
        title = drawing.text(
            self.title, ((self.x+3)*mm, (self.y+10)*mm),
            fill=txtcol, font_size=32,
            font_weight='bold', font_family='sans-serif')
        g.add(title)
        parentgroup.add(g)

    def borderlines(self):
        """liefert die vier linien, die die bounding-box bilden2"""
        # self.x/y ist die linke obere Ecke, +height/width ist dann rechts unten
        return([
            Line(self.x, self.y, self.x+self.width, self.y),         # horiz-oben
            Line(self.x, self.y, self.x, self.y+self.height),         # vert-links
            Line(self.x, self.y+self.height, self.x+self.width, self.y+self.height),   # horiz-unten
            Line(self.x+self.width, self.y, self.x+self.width, self.y+self.height)     # vert-rechts
        ])

    def __str__(self):
        result = "%i %s\n" % (self.id, self.title)
        result += "\t%f/%f/%f, %s x %s, colour=%s" % (
                     self.x, self.y, self.z,
                     self.width, self.height, self.headcolour)
        return(result)

class Notes(list):
    def __init__(self):
        self.by_id = {}

    def addnote(self, datadict, headcolours):
        n = Note(datadict, headcolours)
        self.append(n)
        self.by_id[n.id] = n

    def addall(self, drawing, group):
        """add all notes to a group (sorted by their layers)"""
        # build uniqe list of layers, and sort it
        layerlist = []
        for n in self:
            if n.z not in layerlist:
                layerlist.append(n.z)
        layerlist.sort()
        # traverse it and add every note of the current layer
        for ly in layerlist:
            for n in filter(lambda x: x.z == ly, self):
                n.addto(drawing, group)

    def translate(self, xdelta, ydelta):
        """move all notes contained by x/y-distance"""
        for n in self:
            n.x += xdelta
            n.y += ydelta

    def boundingbox(self):
        """boundingbox over all,
        given by (minimal-x,minimal-y),(maximal-x,maximal-y) over all notes contained"""
        # set "arbitrary "initial-values
        n = self[0]
        xmin = n.x
        xmax = n.x
        ymin = n.y
        ymax = n.y
        # find total min/max values for x/y
        for n in self:
            if xmin > n.x: xmin = n.x
            if xmax < n.x: xmax = n.x
            if ymin > n.y: ymin = n.y
            if ymax < n.y: ymax = n.y
        return(Point(xmin,ymin), Point(xmax+n.width,ymax+n.height))

class Link:
    def __init__(self, datadict, notes):
        """datadict is
        { "dst": 12, "id": 5, "src": 11, "title": "blabla" }
        coordinates will be calculated by fetching notes for src and dst noteID
        and using the center-point of the Note-Widget
        """
        self.id = datadict['id'] # ggf. verwerten
        self.src = notes.by_id[datadict['src']]
        self.dst = notes.by_id[datadict['dst']]
        self.title = datadict['title']
        # calculate start/end as center of src/dst notes
        self.xstart = (self.src.width/2.0) + self.src.x
        self.ystart = (self.src.height/2.0) + self.src.y
        self.xend = (self.dst.width/2.0) + self.dst.x
        self.yend = (self.dst.height/2.0) + self.dst.y
        self.line = Line(self.xstart, self.ystart, self.xend, self.yend)
        isect = self.intersect_dst()
        if isect:
            # Linie auf den Schnittpunkt verkuerzen, damit man den Endmarker
            # (Pfeilspitze) auch sieht.
            if abs(isect.distance(self.line.end)) < abs(isect.distance(self.line.start)):
                self.xend = isect.x
                self.yend = isect.y
            else:
                self.xstart = isect.x
                self.ystart = isect.y

    def translate(self, xdelta, ydelta):
        "translate own graphic-elements to somewhere"
        self.xstart += xdelta
        self.ystart += ydelta
        self.xend += xdelta
        self.yend += ydelta
        self.line = Line(self.xstart, self.ystart, self.xend, self.yend)

    def intersect_dst(self):
        """test, wo die eigene linie eine Begrenzungen  der Zielnotiz schneidet
        die Koordinaten des (ersten) Schnittpunktes zurück, sonst None"""
        for noteborder in self.dst.borderlines():
            isect = self.line.intersect(noteborder)
            if isect:
                return(isect)
        return(None)

    def addto(self, drawing, parentgroup, marker):
        """add self to an svg-drawing"""
        l = drawing.line(
            start=(self.xstart*mm, self.ystart*mm),
            end=(self.xend*mm, self.yend*mm),
            class_='linkbase')
        l.set_markers((None, None, marker))
        parentgroup.add(l)
        if self.title:
            m = self.line.midpoint()
            title = drawing.text(
                        self.title, (m.x*mm, m.y*mm),
                        fill='black',  font_size=32,
                        font_weight='bold', font_family='sans-serif')
            parentgroup.add(title)

    def __str__(self):
        result = "link %i %s\n" % (self.id, self.title)
        return(result)

class Links(list):
    def translate(self, xdelta, ydelta):
        for l in self:
            l.translate(xdelta, ydelta)

    def addall(self, drawing, group, marker):
        """add all Links contained to a group"""
        for l in self:
            l.addto(drawing, group, marker)

class Drawing:
    def __init__(self, notes_links, outfilename):
        (self.notes, self.links) = notes_links
        self.outfilename = outfilename
        # Abmessungen der Zeichnung finden
        (minpoint, maxpoint) = self.notes.boundingbox()
        # todo: svgwrite nach bounding-box f. drawing gucken
        # a4quer, 2*10mm Rand für Papiergröße
        paperwidth = 297
        paperheight = 210
        marginwidth = 10
        # Verschiebung zur linken oberen Ecke des Papiers
        xdelta = -minpoint.x
        ydelta = -minpoint.y
        self.notes.translate(xdelta, ydelta)
        self.links.translate(xdelta, ydelta)
        # Skalierungsfaktor muss die Ränder auf beiden Seiten berücksichtigen
        scale = self.calculate_scaling(
            minpoint, maxpoint,
            paperwidth-2*marginwidth, paperheight-2*marginwidth)

        # Zeichnung bauen und bestuecken
        self.dwg = svgwrite.Drawing(outfilename,
            size=(paperwidth*mm, paperheight*mm), debug=True,
            viewBox=(minpoint.x, minpoint.y, maxpoint.x, maxpoint.y))
        self.init_styles()
        marker = self.init_arrowhead()
        # links und notizen in eine Gruppe packen, die man spaeter skalieren kann
        self.maingroup = self.dwg.g(id='maingroup')
        self.links.addall(self.dwg, self.maingroup, marker)
        self.notes.addall(self.dwg, self.maingroup)
        # alles zusammen skalieren und passend zum Rand verschieben
        margin = marginwidth*3.54
        self.maingroup.translate(margin, margin)
        self.maingroup.scale(scale)
        self.dwg.add(self.maingroup)

    def calculate_scaling(self, minpoint, maxpoint, paperwidth, paperheight):
        # skalierung ausrechnen
        drawing_x = abs(maxpoint.x - minpoint.x)
        drawing_y = abs(maxpoint.y - minpoint.y)
        scale_x = (paperwidth)/(drawing_x)
        scale_y = (paperheight)/(drawing_y)
        result = min(scale_x, scale_y)
        if result == scale_x:
            print ("scale_x")
        else:
            print ("scale_y")
        return(result)

    def init_styles(self):
        """initialize global styles"""
        STYLES = """
        .notebase{fill: white; stroke: black; stroke-width: 3;}
        .headbase{stroke: black; stroke-width: 3;}
        .linkbase{stroke: black; stroke-width: 3;}
        """
        self.dwg.defs.add(self.dwg.style(STYLES))

    def init_arrowhead(self):
        """creates an arrowhead (triangle) marker to defs and return it"""
        # general parameters for arrowhead-marker
        marker = self.dwg.marker(
            insert=(10,10), size=(20,20),
            overflow="visible",
            orient="auto",
            refY="0.0",
            refX="0.0"
        )
        # graphical path for arrowhead-marker (triangle)
        ahead = self.dwg.path(
            d="M0,0 L0,6 L9,3 z",
            fill_rule="evenodd", fill_opacity=1, fill="#000000",
            stroke="#000000", stroke_opacity=1, stroke_width="1pt",
            transform="scale(1.15) translate(0,5.7)"
        )
        marker.add(ahead)
        self.dwg.defs.add(marker)
        return(marker)

    def save(self):
        self.dwg.save(pretty=True)

def readnotes(filename):
    notes = Notes()   # verzauberte Liste(n)
    headcolours = []
    links = Links()
    input = open(filename, 'r', encoding='utf-8')
    input.readline()
    (maxNoteID, maxLinkID, datasets) = jsonpickle.decode(input.read())
    input.close
    for (datatype, data) in datasets:
        if datatype == 'COLOURLABELS':
            # list of [bgcol, txtcol, colourname]
            headcolours = data
        elif datatype == 'NOTES':
            for note in data:
                notes.addnote(note, headcolours)
        elif datatype == 'LINKS':
            for link in data:
                links.append(Link(link, notes))
    return(notes, links)

if __name__ == '__main__':
    if not (len(sys.argv) == 3):
        print ("Export svg-graphics from nzz-file")
        print ("Usage: %s <infile.nzz> <outfile.svg>" % sys.argv[0])
        sys.exit(1)
    d = Drawing(readnotes(sys.argv[1]), sys.argv[2])
    d.save()

# Next:
#
# update Roadmap (Versionszuordnung, Feature svgwrite, distro)
#
# Als Export-Funktion in's Projekt einbauen
#        Menuepunkt "Export/Graphics as SVG" (abkupfern bei headcolour->...)
#        Save-File Dialog dazu
#        Refactoring f. Benutzung der internen Datenstrukturen
#        SVG/Funktionen in separate Lib schreiben
#        Distro bauen
#        alle Rechner updaten

# eintragen bei pypi?
#        Doku sphinx schreiben
#        Distro online stellen
