#!/usr/bin/env python3
#coding:utf-8
# Author:  Wolfram Eifler --<notiznetz@weifler.in-berlin.de>
# Purpose: Qt scene for the GraphicsArea, holds NoteWidgets and LinkWidgets to be displayed in a View
# Created: 01/22/2017
# Copyright 2017, 2021 Wolfram Eifler
# Usage-Permission: GNU General Public License version 3 as published by the
# Free Software Foundation. The full text of the license is availabe
# in this distribution in the file licenses/gpl-v3.txt or
# via internet at http://www.gnu.org/licenses/gpl-3.0.html
"""
Qt scene for the GraphicsArea, holds NoteWidgets and LinkWidgets to be
displayed in a View
"""

import sys, os, time
from PyQt5 import Qt, QtGui, QtWidgets, QtCore
import notiznetz
#from notiznetz.widgets import NoteWidget, LinkWidget, NotesWeb
from notiznetz.util import *

class NoteScene(QtWidgets.QGraphicsScene):
    """for specializing mouse-events"""
    def __init__(self, parent, containing_scene=None, containing_note=None):
        assert(isinstance(parent, notiznetz.mainview.MainView))
        super(NoteScene,self).__init__(parent)
        self.p = parent
        self.data = notiznetz.graph.NotesWeb()
        self.labelcolours = LabelColours()
        # enum-type
        self.modes = ('Edit','NewLink','NewNote')
        self.mode = 'Edit'
        self.openlink =        None        # Link with an open end when creating
        self.subscenes = {}        # holds prepared scenes for notes containing notes
        self.p.statusBar().showMessage(self.mode)
        self.containing_scene = containing_scene # for zoomIn/zoomOut of subnote-maps
        self.containing_note = containing_note        # for setting parent-id

    def isToplevel(self):
        """am i a toplevel-scene or a subscene?"""
        return(self.containing_scene == None)

    def filterNotes(self, objectlist):
        """filters out objects of Type NoteWidget from objectlist"""
        return(filter(lambda x: isinstance(x,notiznetz.widgets.NoteWidget),objectlist))

    def selectedNotes(self):
        return(self.filterNotes(self.selectedItems()))

    def filterLinks(self, objectlist):
        """filters out objects of Type LinkWidget from objectlist"""
        return(filter(lambda x: isinstance(x,notiznetz.widgets.LinkWidget),objectlist))

    def selectedLinks(self):
        return(self.filterLinks(self.selectedItems()))

    def newNote(self):
        n = self.data.newNote()
        if not self.isToplevel():
            n.parentid = self.containing_note.data.ident
        nw = notiznetz.widgets.NoteWidget(n, self)
        # put the new note 200 points left of the current visible-center
        nw.setPos(self.p.visibleCenter() + Qt.QPointF(-200, 0))
        # check for collision in self with other notes with n.collidesWith
        # use n.moveby(dx, dy)
        while list(self.filterNotes(self.collidingItems(nw))):
            nw.moveBy(52, 0)
        for n2 in self.filterNotes(self.selectedItems()):
            n2.setSelected(False)
        nw.setSelected(True)
        self.addItem(nw)
        nw.headlinetext.setEdit(True)
        self.p.signalDataChange()
        #print ('hcol', nw.headlinecolour)
        #print ('subnotes', nw.subnotes)
        return(nw)

    def setHeadlineColour(self, colourindex):
        """sets the headlinecolour of the currently selected Notes to the
        colour indicated by colourindex (which is an integer pointing into
        labelcolours-table)."""
        for nw in self.selectedNotes():
            assert(isinstance(nw, notiznetz.widgets.NoteWidget))
            nw.setHeadlineColour(colourindex)
            self.p.signalDataChange()

    def adjustNoteStacking(self, amount):
        """If amount is positiv raise the currently selected Note(s) in stacking-
        order. If negative lower it"""
        for nw in self.selectedNotes():
            assert(isinstance(nw, notiznetz.widgets.NoteWidget))
            targetval = nw.zValue() + amount
            if targetval < 10:
                nw.setZValue(10)
            elif targetval > 1000:
                nw.setZValue(1000)
            else:
                nw.setZValue(targetval)
            self.p.signalDataChange()

    def deleteLink(self, linkwidget):
        assert(isinstance(linkwidget, notiznetz.widgets.LinkWidget))
        print ("deleting link", linkwidget)
        for nw in self.filterNotes(self.items()):
            assert(isinstance(nw, notiznetz.widgets.NoteWidget))
            newlinks = []
            for lw in nw.linkwidgets:
                if lw != linkwidget:
                    newlinks.append(lw)
            nw.linkwidgets = newlinks
        self.data.deleteLink(linkwidget.data)
        self.removeItem(linkwidget)
        self.p.signalDataChange()

    def deleteNoteAttachedLinks(self, notewidget):
        assert(isinstance(notewidget, notiznetz.widgets.NoteWidget))
        for lw in notewidget.linkwidgets:
            self.deleteLink(lw)

    def deleteNote(self, notewidget):
        assert(isinstance(notewidget, notiznetz.widgets.NoteWidget))
        print ("delete note now: %s" % notewidget.data.title)
        self.data.deleteNoteAttachedLinks(notewidget.data)
        self.deleteNoteAttachedLinks(notewidget)
        nd = notewidget.data
        self.removeItem(notewidget)
        # delete note from data
        self.data.deleteNote(nd)
        self.p.signalDataChange()

    def subScene(self, notewidget):
        """returns the scene for all notes contained in note,
        creates it if necessary.
        """
        assert(isinstance(notewidget, notiznetz.widgets.NoteWidget))
        if notewidget.data.ident in self.subscenes:
            return(self.subscenes[notewidget.data.ident])
        result = NoteScene(self.p, self, notewidget)
        # IMPORTANT, otherwise subscene-data is lost when saving
        result.data = self.data
        # add corresponding notewidgets to the scene
        for nw in notewidget.subnotes:
            result.addItem(nw)
        # todo: find links of note, find corresponding linkwidget
        # add it to the scene. currently links for subnotes are note even loaded
        #
        notewidget.subscene = result
        self.subscenes[notewidget.data.ident] = result
        return(result)

    def zoomIntoNote(self, notewidget):
        """zooms into the card-space of notewidget thus making its subnotes
        visible"""
        assert(isinstance(notewidget, notiznetz.widgets.NoteWidget))
        return()    # still not ready in 0.0.3
        if len(notewidget.subnotes):
            print ("ctrl-down: zoom-into submap of '%s' now" % notewidget.data.title)
            s = notewidget.subscene
            for nw in notewidget.subnotes:
                nw.scene = s
            self.p.scene = s
            self.p.graphicsView.setScene(s)
            self.p.addCrosshair(0,0,10)

    def zoomOutOfNote(self):
        """zooms out of current-notes cardspace if not at toplevel"""
        return()    # still not ready in 0.0.3
        if not self.isToplevel():
            print ("ctrl-up and in subcene: zooming out of submap.")
            self.p.scene = self.containing_scene
            self.p.graphicsView.setScene(self.containing_scene)

    def setMode(self, modename):
        if self.modes.count(modename):
            self.mode = modename
            self.p.statusBar().showMessage(modename)

    def mouseMoveEvent(self, e):
        QtWidgets.QGraphicsScene.mouseMoveEvent(self, e)
        if self.mode == 'NewLink' and self.openlink:
            # in NewLink mode: update link-endpoint to mousecursor-point
            assert(isinstance(self.openlink, notiznetz.widgets.LinkWidget))
            self.openlink.adjust(e.scenePos())

    def mousePressEvent(self, e):
        # click outside the note-widget if editing headline commits edit-changes
        assert(isinstance(e, QtWidgets.QGraphicsSceneMouseEvent))
        if e.modifiers() == QtCore.Qt.ShiftModifier:
            g = self.p.graphicsView
            g.setDragMode(g.ScrollHandDrag)
        if ((e.button() == QtCore.Qt.LeftButton)):
            for n in self.filterNotes(self.items()):
                if n.headlinetext.editing:
                    n.headlinetext.commitEdit()
            for l in self.filterLinks(self.items()):
                if l.titletext.editing:
                    l.titletext.commitEdit()
        if e.modifiers() == QtCore.Qt.ControlModifier:
            if e.button() == QtCore.Qt.MiddleButton:
                self.p.zoomSetStd()
        QtWidgets.QGraphicsScene.mousePressEvent(self, e)

    def mouseReleaseEvent(self, e):
        g = self.p.graphicsView
        g.setDragMode(g.RubberBandDrag)
        QtWidgets.QGraphicsScene.mouseReleaseEvent(self, e)

    def wheelEvent(self, e):
        if e.modifiers() == QtCore.Qt.ControlModifier:
            if e.delta() < 0:
                self.p.zoomIn()
            else:
                self.p.zoomOut()
        else:
            e.ignore()

    def keyPressEvent(self, e):
        """implementation of
        add-note on shift-returnkey,
        ctrl-Q (Quit), ctrl-space (edit notes-text),
        del- (delete note/link) and esc-key
        ctrl-arrowdown, ctrl-arrowup for zooming into submaps for contained-notes
        """
        k = e.key()
        if e.modifiers() == QtCore.Qt.ShiftModifier:
            if (k == QtCore.Qt.Key_Return) or (k == QtCore.Qt.Key_Enter):
                self.newNote()
            else:
                QtWidgets.QGraphicsScene.keyPressEvent(self, e)
        elif (e.modifiers() == QtCore.Qt.ControlModifier):
            if (k == QtCore.Qt.Key_Q):
                self.p.quitProgram()
            elif (k == QtCore.Qt.Key_E) or (k == QtCore.Qt.Key_Space):
                self.p.editSelectedNotes()
            elif (k == QtCore.Qt.Key_Down):
                l = self.selectedNotes()
                if len(l) == 1:
                    self.zoomIntoNote(l[0])
            elif(k == QtCore.Qt.Key_Up):
                self.zoomOutOfNote()
        elif (k == QtCore.Qt.Key_Delete):
            for note in self.selectedNotes():
                self.deleteNote(note)
            for link in self.selectedLinks():
                self.deleteLink(link)
        elif (k == QtCore.Qt.Key_Escape) and (self.mode == 'NewLink'):
            self.removeItem(self.openlink)
            self.openlink = None
            self.setMode('Edit')
        else:
            QtWidgets.QGraphicsScene.keyPressEvent(self, e)

    def clear(self):
        self.data = notiznetz.graph.NotesWeb()
        super(NoteScene,self).clear()
        self.p.addCrosshair(0,0,10)

    def load(self, filename):
        self.clear()
        mainwindata = self.data.load(filename)
        self.labelcolours = self.data.labelcolours
        subnotes = {}
        allnotes = {}
        alllinks = {}
        toplevelnotes = []
        # first round: load all object-data and make widgets for toplevel-notes
        for n in self.data.getAllNotes():
            assert(isinstance(n, notiznetz.graph.Note))
            nw = notiznetz.widgets.NoteWidget(n, self) # make a widget
            if n.editwin[0]:        # visible-flag
                self.p.openTextEdit(nw)
            allnotes[nw.data.ident] = nw
            if n.parentid == 0:        # only for toplevelnotes...
                toplevelnotes.append(nw)
                self.addItem(nw)
            else:
                subnotes[nw.data.ident] = nw
        for l in self.data.getAllLinks():
            # find  corresponding src/dst note-widgets, link has only
            # note objects
            src  = None
            dst = None
            for nw in allnotes.values():
                if nw.data.ident == l.src.ident:
                    src = nw
                elif nw.data.ident == l.dst.ident:
                    dst = nw
            if (src != None) and (dst != None):
                lw = notiznetz.widgets.LinkWidget(l, src)
                lw.connect(src, dst)
                alllinks[lw.data.ident] = lw
                # only links between toplevelnotes
                if ((src.data.parentid == dst.data.parentid) and src.data.parentid == 0):
                    self.addItem(lw)
        # second round: add childnotes to parents
        for nw in subnotes.values():
            parentnote = allnotes[nw.data.parentid]
            assert(isinstance(parentnote, notiznetz.widgets.NoteWidget))
            parentnote.addNote(nw)
        # third round: adjust positions of the widgets according to loaded data
        # and create subnote-scenes
        for nw in allnotes.values():
            (x,y,width,height) = nw.data.getGeometry()
            nw.moveBy(x+(width/2), y+(height/2))
            if len(nw.subnotes):
                self.subScene(nw)
        # Add Links  of subnotes to corresponding subnote-scenes
        for lw in alllinks.values():
            srcparentid = lw.src.data.parentid
            dstparentid = lw.dst.data.parentid
            if (srcparentid == dstparentid) and (srcparentid != 0):
                parentnote = allnotes[srcparentid]
                parentnote.subscene.addItem(lw)
                lw.adjust()
        for lw in self.filterLinks(self.items()):
            lw.adjust()
        return(mainwindata)

    def syncNotes(self, scene):
        assert(isinstance(scene, NoteScene))
        for nw in self.filterNotes(scene.items()):
            assert(isinstance(nw, notiznetz.widgets.NoteWidget))
            pos = nw.mapToScene(-50, -25)
            #print (nw.data.title, pos)
            nw.data.setGeometry(pos.x(), pos.y(), z=nw.zValue())
            ew = nw.editwin
            # [bool:visible, x,y, width,height, scrollpos]
            if ew:
                ewdata = [
                        ew.isVisible(),
                        ew.x(), ew.y(), ew.width(), ew.height(),
                        ew.editor.verticalScrollBar().value()
                    ]
            else:
                ewdata = [False, 100,100, 500,250, 0]

            nw.data.setEditWindow(ewdata)

    def syncNotesSubnotes(self):
        self.syncNotes(self)
        for ssc in self.subscenes.values():
            self.syncNotes(ssc)

    def save(self, filename):
        if not self.isToplevel():
            return(self.p.toplevel_scene.save(filename))
        # first synchronize geometries etc of noteswidgets to notes
        self.syncNotesSubnotes()
        self.data.save(filename, self.p.getScreenPosition())

    def export_svg(self, filename):
        if not self.isToplevel():
            return(self.p.toplevel_scene.export_svg(filename))
        # first synchronize geometries etc of noteswidgets to notes
        self.syncNotesSubnotes()
        self.data.export_svg(filename, self.p.getScreenPosition())
