#!/usr/bin/env python3
#coding:utf-8
# Author:  Wolfram Eifler --<notiznetz@weifler.in-berlin.de>
# Purpose: Persistencelayer to disk for notesweb
# Created: 01/27/2017
# Copyright 2017, 2021 Wolfram Eifler
# Usage-Permission: GNU General Public License version 3 as published by the
# Free Software Foundation. The full text of the license is availabe
# in this distribution in the file licenses/gpl-v3.txt or
# via internet at http://www.gnu.org/licenses/gpl-3.0.html
"""
Persistencelayer to disk for notesweb
"""

# TODO: Farben wurden erweitert, jetzt schmiert er ab, wenn in einer alten Datei eine neue Farbe
# vergeben wird, weil die Liste in util.py aus der Datei initialisiert wurde.
# Ansatz: Liste bekannter Farben gegen Farbliste aus Datei abgleichen und ggf. stillschweigend ergaenzen

# todo-later: Konsistenzbedingungen/Selbsttest implementieren

import os, jsonpickle
#from data import NoteData, LinkData
import notiznetz
import notiznetz.graph
from notiznetz.util import *

fmtline_prefix = '# application/notiznetz'

# set prettyprinting for json-encoders so sourcecode-controlsystems can
# do sensible diffs on data-files
jsonpickle.set_encoder_options('simplejson', sort_keys=True, indent=2)
jsonpickle.set_encoder_options('json', sort_keys=True, indent=2)
jsonpickle.set_encoder_options('demjson', compactly=False)

def detect_fileformat(filename):
    # check the version-tag
    input = open(filename, 'r')
    line = input.readline()
    input.close()
    if line.startswith(fmtline_prefix):
        fmt = line.strip().split()[2]
        return(fmt)
    print (line.strip())
    return('unknown')

#def linklists_equal(links1, links2):
    #result = True
    #if len(links1) != len(links2):
        #result = False
    #else:
        #for i in range(0, len(links1)):
            #l1 = links1[i]; assert(isinstance(l1, LinkData))
            #l2 = links2[i]; assert(isinstance(l2, LinkData))
            #if not l1.equal(l2): result = False
    #return(result)

def dataToNote(notedata):
    """create a Note-Object for a graph from data (read from disk)
    notedata is a dictionary"""
    assert(isinstance(notedata, dict))
    if not ('parentid' in notedata):
        notedata['parentid'] = 0
    n = notiznetz.graph.Note(notedata['id'], parentid=notedata['parentid'],
                   title=notedata['title'], text=notedata['text'])
    n.setGeometry(notedata['x'], notedata['y'],
                  notedata['width'], notedata['height'], notedata['z'])
    # only from version 0.0.3 on
    if 'headcolour' in notedata:
        n.setHeadlineColour(notedata['headcolour'])
    if 'edit' in notedata:
        ew = notedata['edit']  # data for editwindow
        n.setEditWindow([
            ew['visible'],
            ew['x'],ew['y'], ew['width'],ew['height'],
            ew['scrollpos']
        ])
    return(n)

def noteToData(note):
    """create a data-object suitable for writing to disk
    from a Note-Object (used in the graph)"""
    assert(isinstance(note,notiznetz.graph.Note))
    (visible, editx, edity, editwidth, editheight, scrollpos) = note.editwin
    nd = {
        'id' : note.ident,
        'parentid' : note.parentid,
        'headcolour' : note.headlinecolour,
        'title' : note.title,
        'text' : note.text,
        'edit' : {
            'visible' : visible,
            'x' : editx,
            'y' : edity,
            'width' : editwidth,
            'height' : editheight,
            'scrollpos' : scrollpos
            },
        'x' : note.x,
        'y' : note.y,
        'z' : note.z,
        'width' : note.width,
        'height' : note.height
    }
    return(nd)

def linkToData(link):
    """"""
    assert(isinstance(link, notiznetz.graph.Link))
    ld = {
        'id' : link.ident,
        'title' : link.title,
        'src' : link.src.ident,
        'dst' : link.dst.ident
    }
    return(ld)

def dataToLink(linkdata, notesdict):
    """create a Link-Object for a graph from link-data (read from disk)
    notesdict is a dictionary of ident->Note-Object, used to find the
    starting- and ending-Note of the link. notesdata is a dict
    fieldname->fieldvalue"""
    assert(isinstance(linkdata, dict))
    srcnote = notesdict[linkdata['src']]
    dstnote = notesdict[linkdata['dst']]
    l = notiznetz.graph.Link(linkdata['id'], srcnote, dstnote, linkdata['title'])
    return(l)

class Store:
    # to be changed dont need NotesWeb for now
    def __init__(self):
        self.maxNoteID = 0
        self.maxLinkID = 0
        self.links = {}
        self.notes = {}
        self.labelcolours = LabelColours()
        self.mainWinData = []
        # dispatcher-table for loading different fileformats,
        # formatversion -> loading-function
        self.known_fileformats = {
        '0.0.2' : self._load_0_0_2,
        '0.0.2-testing' : self._load_0_0_2,
        '0.0.3' : self._load_0_0_3,
        '0.0.3-testing' : self._load_0_0_3,
        }

    def putNoteID(self, idcounter):
        assert(isinstance(idcounter, IdentCounter))
        self.maxNoteID = idcounter._getCurrent()

    def putLinkID(self, idcounter):
        assert(isinstance(idcounter, IdentCounter))
        self.maxLinkID = idcounter._getCurrent()

    def putMainwinData(self, mainwindata):
        """geometry-data of main-window:
        [x, y, width, height, scrollx, scrolly, zoomfactor]"""
        self.mainWinData = mainwindata

    def addNote(self, note):
        assert(isinstance(note,  notiznetz.graph.Note))
        self.notes[note.ident] = note
        if self.maxNoteID < note.ident:
            self.maxNoteID = note.ident

    def addLink(self, link):
        assert(isinstance(link, notiznetz.graph.Link))
        # link has to have src and dst, and src and dst note
        # have to be already in store
        assert(link.src)
        assert(link.dst)
        assert(link.src.ident in self.notes)
        assert(link.dst.ident in self.notes)
        self.links[link.ident] = link
        if self.maxLinkID < link.ident:
            self.maxLinkID = link.ident

    def getAllNotes(self):
        return(self.notes.values())

    def getAllLinks(self):
        return(self.links.values())

    def load(self, filename):
        ffmt = detect_fileformat(filename)
        if ffmt in self.known_fileformats:
            try:
                self.known_fileformats[ffmt](filename)
                if not self.labelcolours:
                    self.labelcolours = LabelColours()
                if not self.mainWinData:
                    self.mainWinData = [0,0, 600,450, 0, 0, 1.0]
            except(ValueError):
                sys.stderr.write(
                "ERROR: in decoding fileformat %s, '%s' not loaded.\n" %
                (ffmt, filename))
        else:
            sys.stderr.write("ERROR: can not handle fileformat '%s'\n" % ffmt)

    def _load_0_0_2(self, filename):
        print ("loading format 0.0.2")
        input = open(filename, 'rb')
        input.readline()                # version-line
        (self.maxNoteID, self.maxLinkID, notesdata, linksdata
             ) = jsonpickle.decode(input.read())
        input.close()
        # convert data-objects to note/link-objects suitable for the graph
        for notedata in notesdata:
            # set default-values used in old program-version
            if not ('z' in notedata):
#                notedata['z'] = 97
                notedata['z'] = 93  # for testing if z is really loaded
            self.notes[notedata['id']] = dataToNote(notedata)
        for linkdata in linksdata:
            self.links[linkdata['id']] = dataToLink(linkdata, self.notes)

    def _load_0_0_3(self, filename):
        print ("loading format 0.0.3")
        input = open(filename, 'rb')
        input.readline()                # version-line
        (self.maxNoteID, self.maxLinkID, datasets
             ) = jsonpickle.decode(input.read())
        input.close()
        mainwindata = None
        notesdata = None
        linksdata = None
        lablesdata = None
        # [colourlabels, [colourdef, labeltext], ...]
        for (datatype, data) in datasets:
            if datatype == 'MAINWINDOW':
                self.mainWinData = data
            elif datatype == 'NOTES':
                notesdata = data
            elif datatype == 'LINKS':
                linksdata = data
            elif datatype == 'COLOURLABELS':
                lablesdata = data
            else:
                sys.stderr.write("Warning: unknown datatype in file - ignored")
        if not notesdata:
            sys.stderr.write("Warning: no data for notes found")
        if not linksdata:
            sys.stderr.write("Warning: no data for links found")
        if not lablesdata:
            sys.stderr.write("Warning: no data for notecolour/labeltext found")
        # convert data-objects to note/link-objects suitable for the graph
        for notedata in notesdata:
            self.notes[notedata['id']] = dataToNote(notedata)
        for linkdata in linksdata:
            self.links[linkdata['id']] = dataToLink(linkdata, self.notes)
        # put labelsdata to LabelColours-object
        self.labelcolours = LabelColours(lablesdata)
        #print ('store: headlinecolours:', self.labelcolours)

    def jsonsstring(self):
        fmtline = "%s %s\n" % (fmtline_prefix, notiznetz.fileformat_version)
        # convert note/link-objects to data-objects suitable for disk-write
        notesdata = []
        for (ident, note) in self.notes.items():
            notesdata.append(noteToData(note))
        linksdata = []
        for (ident, link) in self.links.items():
            linksdata.append(linkToData(link))
        labelsdata = self.labelcolours.getPlainData()
        datasets = [['MAINWINDOW', self.mainWinData],
                    ['COLOURLABELS', labelsdata],
                    ['NOTES', notesdata],
                    ['LINKS', linksdata]]
        l = [self.maxNoteID, self.maxLinkID, datasets]
        return (fmtline + jsonpickle.encode(l))

    def save(self, filename):
        print ("saving to %s" % filename)
        output = open(filename, 'w', encoding='utf-8')
        output.write(self.jsonsstring())
        output.close()

    #def equals(self, otherstore):
        #"""compare stores value by value"""
        #assert(isinstance(otherstore, Store))
        #result = True
        #if self.maxLinkID != otherstore.maxLinkID: result = False
        #if self.maxNoteID != otherstore.maxNoteID: result = False
        #for n in self.getAllNotes():
            #n2 = otherstore.getNoteData(n.ident)
            #if not n.equals(n2): result = False
        #if not linklists_equal(self.getAllLinks(), otherstore.getAllLinks()):
            #result = False
        #return(result)

if __name__ == '__main__':
    pass
