#!/usr/bin/env python3
#coding:utf-8
# Author:  Wolfram Eifler --<notiznetz@weifler.in-berlin.de>
# Purpose: general-purpose items for notiznetz-application
# Created: 01/23/2017
# Copyright 2017, Wolfram Eifler
# Usage-Permission: GNU General Public License version 3 as published by the
# Free Software Foundation. The full text of the license is availabe
# in this distribution in the file licenses/gpl-v3.txt or
# via internet at http://www.gnu.org/licenses/gpl-3.0.html
"""
general-purpose items for notiznetz-application
"""

import sys
import unittest
from PyQt5 import Qt, QtWidgets, QtCore

# pen for drawing the headline-background of a NoteWidget
pen_b14 = Qt.QPen(QtCore.Qt.black)
pen_b14.setWidth(14)
pen_b0 = Qt.QPen(QtCore.Qt.black)
pen_b0.setWidth(0)
pen_gray = Qt.QPen(QtCore.Qt.gray)
#
brush_white = Qt.QBrush(QtCore.Qt.white)
brush_black = Qt.QBrush(QtCore.Qt.black)
brush_yellow = Qt.QBrush(QtCore.Qt.yellow)
brush_gray = Qt.QBrush(QtCore.Qt.gray)

def hexcol_to_qcol(hexcol):
    """changes a string of form 'ff00ff' (three hexacecimals)
    to a Qt-Color"""
    if hexcol.startswith('0x'): hexcol = hexcol[2:]
    assert(len(hexcol) == 6)
    return(Qt.QColor(
        int(hexcol[:2], 16), int(hexcol[2:4], 16), int(hexcol[4:], 16)))

def qcol_to_hexcol(qcolour):
    """changes a Qt-Color to a hex-string of form 'ff00ff'
    (three hexacecimals)"""
    assert(isinstance(qcolour, Qt.QColor))
    return(qcolour.name()[1:])

class IdentCounter:
    """for generating internal id's for objects and widgets"""
    def __init__(self):
        self.__current_value__ = 0

    def getID(self):
        self.__current_value__ += 1
        return(self.__current_value__)

    def _getCurrent(self):
        return(self.__current_value__)

    def _setCurrent(self, value):
        """sets the value of the last id given out"""
        assert(type(value) == type(1))
        self.__current_value__ = value

class LabelColours(list):
    """Helper for indexed label-colour-entries, which are of form
    [backgroundcolour, textcolour, labeltext],
    lookup is done via standard list-indexing
    when appending the hex-values backgroundcolor and textcolour are converted
    to qt-brush-compatible qt-colours. To get the plain-values back use
    getPlainData, it returns a python list of [[hex, hex, string],...]"""

    def __init__(self, items=[]):
        super(LabelColours, self).__init__()
        black = Qt.QColor(QtCore.Qt.black)
        white = Qt.QColor(QtCore.Qt.white)
        darkblue = Qt.QColor(QtCore.Qt.darkBlue)
        lightgreen = Qt.QColor(QtCore.Qt.green)
        darkgreen = Qt.QColor(QtCore.Qt.darkGreen)
        darkgray = Qt.QColor(QtCore.Qt.darkGray)
        red = Qt.QColor(QtCore.Qt.red)
        yellow = Qt.QColor(QtCore.Qt.yellow)
        cyan = Qt.QColor(QtCore.Qt.cyan)
        darkcyan = Qt.QColor(QtCore.Qt.darkCyan)
        darkred = Qt.QColor(QtCore.Qt.darkRed)
        stdlist = []
        stdlist.append([black, white, 'black'])
        stdlist.append([darkgreen, white, 'green'])
        stdlist.append([darkblue, white, 'blue'])
        stdlist.append([darkgray, white, 'grey'])
        stdlist.append([red, white, 'red'])
        stdlist.append([yellow, black, 'yellow'])
        stdlist.append([darkred, white, 'darkred'])
        stdlist.append([lightgreen, black, 'neongreen'])
        stdlist.append([cyan, black, 'cyan'])
        stdlist.append([darkcyan, white, 'darkcyan'])
        if not items:
            for elem in stdlist:
                self.append(elem)
        else:
            # different colourlist (maybe a shorter one from an old file) use that one first...
            for item in items:
                assert(len(item) == 3)
                (bgcol, txtcol, txt) = item
                # creating qt-colours here
                self.append([hexcol_to_qcol(bgcol),hexcol_to_qcol(txtcol), txt])
            # and append the missing ones from stdlist
            for elem in stdlist:
                if not elem in self:
                    self.append(elem)

    def getPlainData(self):
        """for storage-format"""
        # todo: implement dummy, implement back-conversion qt-color-to-hex
        result = []
        for (bgcol, txtcol, txt) in self:
            result.append([qcol_to_hexcol(bgcol), qcol_to_hexcol(txtcol), txt])
        return(result)

if __name__=='__main__':
    unittest.main()
