#!/usr/bin/env python3
#coding:utf-8
# Author:  Wolfram Eifler --<notiznetz@weifler.in-berlin.de>
# Purpose: Widgets for a Notes and Links between them
# Created: 01/23/2017
# Copyright 2017, 2021 Wolfram Eifler
# Usage-Permission: GNU General Public License version 3 as published by the
# Free Software Foundation. The full text of the license is availabe
# in this distribution in the file licenses/gpl-v3.txt or
# via internet at http://www.gnu.org/licenses/gpl-3.0.html
"""
Widgets for a Notes and Links between them
"""

import sys
import unittest
from PyQt5 import Qt, QtWidgets, QtCore, QtGui
import notiznetz
from notiznetz.graph import Note, Link
from notiznetz.util import *

class HeadlineText(Qt.QGraphicsTextItem):
    """headline of a Note, subclassed for editing-functions of the note-title"""
    def __init__(self, x, y, text, parent):
        assert(isinstance(parent, NoteWidget))
        super(HeadlineText, self).__init__(text, parent)
        self.p = parent
        self.setDefaultTextColor(parent.headlinecolour[1])
        self.setPos(x,y)
        #self.setPos(-40, -30)
        self.setZValue(99)
        self.undo_text = text
        self.editing = False
        self.setToolTip(text)

    def selectAllText(self):
        c = self.textCursor()
        c.select(Qt.QTextCursor.Document)
        self.setTextCursor(c)

    def deselectAllText(self):
        c = self.textCursor()
        c.clearSelection()
        self.setTextCursor(c)

    def setEdit(self, flag):
        """set/unset editing-mode"""
        if flag:
            self.setTextInteractionFlags(QtCore.Qt.TextEditable)
            # activate editing and preselect all text
            self.setFocus(True)
            self.selectAllText()
        else:
            self.deselectAllText()
            self.clearFocus()
            self.setTextInteractionFlags(QtCore.Qt.NoTextInteraction)
            self.update()
        self.editing = flag

    def commitEdit(self):
        # update undo_text and let parent transfer changed text to data-model
        txtplain = self.toPlainText()
        self.undo_text = txtplain
        self.p.setTitle(txtplain)
        self.setEdit(False)
        self.setToolTip(txtplain)
        self.p.scene.p.signalDataChange()

    def abortEdit(self):
        self.setPlainText(self.undo_text)
        self.setEdit(False)

    def keyPressEvent(self, e):
        """implementation of special actions on return, enter and escape"""
        assert(isinstance(e, QtGui.QKeyEvent))
        k = e.key()
        if (k == QtCore.Qt.Key_Return) or (k == QtCore.Qt.Key_Enter):
            self.commitEdit()
        elif k == QtCore.Qt.Key_Escape:
            self.abortEdit()
        else:
            Qt.QGraphicsTextItem.keyPressEvent(self, e)

class NoteHeadline(QtWidgets.QGraphicsRectItem):
    def __init__(self, x, y, width, height, parent=None):
        super(NoteHeadline, self).__init__(x, y, width, height, parent)
        self.p = parent
        self.setZValue(98)
        # use menuHeadcolour from MainView as context-menu
        self.contextmenu = self.p.scene.p.menuContextNoteHeadline

    def contextMenuEvent(self, e):
        # select corresponding NoteWidget so actions of menuHeadcolour will
        # change the colour of this note
        self.p.setSelected(True)
        self.contextmenu.exec_(e.screenPos())

class LinkTitleText(Qt.QGraphicsTextItem):
    """headline of a Note, subclassed for editing-functions of the note-title"""
    def __init__(self, text, parent):
        assert(isinstance(parent, LinkWidget))
        super(LinkTitleText, self).__init__(text, parent)
        self.p = parent
        self.undo_text = text
        self.editing = False

    def adjustPosition(self):
        """adjust own position in middle of line"""
        midline = self.p.line().pointAt(0.5)
        self.setPos(midline)

    def selectAllText(self):
        c = self.textCursor()
        c.select(Qt.QTextCursor.Document)
        self.setTextCursor(c)

    def deselectAllText(self):
        c = self.textCursor()
        c.clearSelection()
        self.setTextCursor(c)

    def setEdit(self, flag):
        """set/unset editing-mode"""
        if flag:
            self.setTextInteractionFlags(QtCore.Qt.TextEditable)
            # activate editing and preselect all text
            self.setFocus(True)
            self.selectAllText()
        else:
            self.deselectAllText()
            self.clearFocus()
            self.setTextInteractionFlags(QtCore.Qt.NoTextInteraction)
            self.update()
        self.editing = flag

    def commitEdit(self):
        # update undo_text and let parent transfer changed text to data-model
        txtplain = self.toPlainText()
        self.undo_text = txtplain
        self.p.setTitle(txtplain)
        self.setEdit(False)
        self.p.scene.p.signalDataChange()

    def abortEdit(self):
        self.setPlainText(self.undo_text)
        self.setEdit(False)

    def mousePressEvent(self, e):
        print (e)
        if (e.button() == QtCore.Qt.LeftButton):
            if (e.modifiers() == QtCore.Qt.ControlModifier):
                self.setEdit(True)

    def keyPressEvent(self, e):
        """implementation of special actions on return, enter and escape"""
        assert(isinstance(e, QtGui.QKeyEvent))
        k = e.key()
        if (k == QtCore.Qt.Key_Return) or (k == QtCore.Qt.Key_Enter):
            self.commitEdit()
        elif k == QtCore.Qt.Key_Escape:
            self.abortEdit()
        else:
            Qt.QGraphicsTextItem.keyPressEvent(self, e)

class LinkArrow(Qt.QGraphicsPolygonItem):
    """arrowhead of a link-line"""
    def __init__(self, parent):
        assert(isinstance(parent, LinkWidget))
        super(LinkArrow, self).__init__(parent)
        self.p = parent
        self.setPen(pen_b0)
        self.setBrush(brush_black)
        self.triangle = Qt.QPolygonF()
        #self.triangle.append(Qt.QPointF(-10, -10))
        #self.triangle.append(Qt.QPointF(3, -3))
        #self.triangle.append(Qt.QPointF(-3, 3))
        self.triangle.append(Qt.QPointF(-5, -4))
        self.triangle.append(Qt.QPointF(-5, 4))
        self.triangle.append(Qt.QPointF(10, 0))
        self.setPolygon(self.triangle)

    def adjust(self):
        l = self.p.line()
        # Versuch, den genauen Schnittpunkt von linie zu Rechteck zu ermittlen,
        # stimmt aber irgendwie nicht (falsche Werte)
        # vermutlich befinden sich Arrow und note in verschiedenen Koordinatensystemen.
        #i = l.NoIntersection
        #ip = Qt.QPointF()
        #rect = self.p.dst.sceneBoundingRect()
        #rect = self.mapRectFromScene(rect)
        #lines_totest = [
            #Qt.QLineF(rect.bottomLeft(), rect.bottomRight()), # bottomline
            #Qt.QLineF(rect.topLeft(), rect.topRight()), # topline
            #Qt.QLineF(rect.bottomLeft(), rect.topLeft()), # leftside
            #Qt.QLineF(rect.bottomRight(), rect.topRight()), # rightside
        #]
        #for line_to_test in lines_totest:
            #i = l.intersect(line_to_test, ip)
            #if i != l.NoIntersection:
                #break
        #print (ip.x(), ip.y())
        #arrowpos = ip
        arrowpos = l.pointAt(0.7)
        self.setPos(arrowpos)
        self.setRotation(-l.angle())

    def mousePressEvent(self, e):
        self.p.mousePressEvent(e)

class LinkWidget(QtWidgets.QGraphicsLineItem):
    """its an edge in the graph, representing a directed named link
       between nodes"""
    def __init__(self, link, startnotewidget):
        assert(isinstance(link, Link))
        assert(isinstance(startnotewidget, NoteWidget))
        super(LinkWidget,self).__init__()
        self.data = link
        #link.title = "this is a debugging test"
        self.setPen(pen_b0)
        self.titletext = LinkTitleText(self.data.title, self)
        #Qt.QGraphicsTextItem(self.data.title, self)
        self.setFlag(self.ItemIsSelectable, enabled=True)
        self.src = startnotewidget
        self.dst = None
        self.scene = startnotewidget.scene
        self.start = None
        self.end = None
        if link.src and link.dst:
            self.data.connect(link.src, link.dst)
        if startnotewidget:
            self.start = startnotewidget.scenePos()
            self.data.src = startnotewidget.data
        self.arrowhead = LinkArrow(self)

    def connect(self, srcnote, dstnote):
        assert(isinstance(srcnote, NoteWidget))
        assert(isinstance(dstnote, NoteWidget))
        if srcnote != dstnote:
            self.src = srcnote
            self.dst = dstnote
            self.src.addLink(self)
            self.dst.addLink(self)
            self.data.connect(self.src.data, self.dst.data)
            self.adjust()

    def adjust(self, endpos=None):
        """updates start/endpoint of the line, which results in a
           GeometryChange thus changing the own bounding-box"""
        self.start = self.src.scenePos()
        xoffset = self.src.data.width/2
        yoffset = self.src.data.height/2
        self.start.setX(self.start.x()+xoffset)
        self.start.setY(self.start.y()+yoffset)
        if self.dst:
            self.end = self.dst.scenePos()
            xoffset = self.dst.data.width/2
            yoffset = self.dst.data.height/2
            self.end.setX(self.end.x()+xoffset)
            self.end.setY(self.end.y()+yoffset)
            self.setLine(QtCore.QLineF(self.start, self.end))
            self.titletext.adjustPosition()
            self.arrowhead.adjust()
        elif (isinstance(endpos, QtCore.QPointF)):
            # no destination-note, this is an open-ended Link (creating-link)
            # use mouse-position for endpoint
            self.setLine(QtCore.QLineF(self.start, endpos))
        else:
            # this is wrong
            print ("Warning: LinkWidget.adjust without endpoint, ignoring")

    def setTitle(self, text):
        self.data.title = text

    def mousePressEvent(self, e):
        assert(isinstance(e, QtWidgets.QGraphicsSceneMouseEvent))
        if (e.button() == QtCore.Qt.LeftButton):
            if (e.modifiers() == QtCore.Qt.ControlModifier):
                # todo: enable link-title-editing here
                self.titletext.setEdit(True)
            else:
                if (self.scene.mode == 'Edit'):
                    self.setSelected(True)
        QtWidgets.QGraphicsLineItem.mousePressEvent(self, e)

    def __str__(self):
        result = "LinkWidget %i:\n" % self.data.ident
        if self.start:
            result += "  from %05.1f / %05.1f\n" % (self.start.x(), self.start.y())
        if self.end:
            result += "  to   %05.1f / %05.1f" % (self.end.x(), self.end.y())
        return(result)


class NoteWidget(QtWidgets.QGraphicsRectItem):
    """its a node in the graph, containing a textnote"""
    def __init__(self, note, scene=None):
        assert(isinstance(note, Note))
        if scene: assert(isinstance(scene,notiznetz.scene.NoteScene))
        super(NoteWidget,self).__init__(0,0,note.width,note.height)
        self.data=note
        self.scene = scene
        self.editwin = None
        self.setFlag(self.ItemIsMovable, enabled=True)
        self.setFlag(self.ItemIsSelectable, enabled=True)
        self.setFlag(self.ItemClipsToShape, enabled=True)
        self.setFlag(self.ItemClipsChildrenToShape, enabled=True)
        #self.setAcceptDrops(True)                # scheint keinen Unterschied zu machen, die Events kommen nicht an
        #self.setCursor(QtCore.Qt.CrossCursor)   # funktioniert!
        self.setBrush(brush_white)
        self.setZValue(note.z)
        #self.setZValue(97)
        # rect for headline
        hx  = 1
        hy = 1
        hwidth = note.width -2
        hheight  = (note.height/5)+4
        self.headlinecolour = self.scene.data.labelcolours[note.headlinecolour]
        #self.headline = QtWidgets.QGraphicsRectItem(-49,-24,98,14, self)
        self.headline = NoteHeadline(hx, hy, hwidth, hheight, self)
        self.headlinetext = HeadlineText(hx+9, hy-6, self.data.title, self)
        self.setHeadlineColour(note.headlinecolour)
        # contentmarker is drawn when a note has text-content
        cmsize = hheight * 0.6
        self.contentmarker = QtWidgets.QGraphicsEllipseItem(
            hx+1, hy+1, cmsize, cmsize, self)
        self.subnotes = []
        self.subscene = None
        self.linkwidgets = []
        self.contentmarker.setBrush(brush_gray)
        self.contentmarker.setPen(pen_gray)
        self.updateContentMarker()

    # future experiments for getting rid of notewidget.data.bla in favour
    # of notewidget.bla. getattr seems to work,
    # setattr does not seem to be called
    #def __getattribute__(self, attrname):
        #if attrname == 'headlinecolour':
            #return(self.data.headlinecolour)
        #else:
            #return(super(NoteWidget,self).__getattribute__(attrname))

    #def __setattribute__(self, attrname, attrvalue):
        #print ('name', attrname)
        #print ('val', attrvalue)

    def updateContentMarker(self):
        if self.data.text or len(self.subnotes):
            # make contentmarker visible
            self.contentmarker.setZValue(99)
        else:
            # hide it below headline-rect
            self.contentmarker.setZValue(0)

    def setText(self, text):
        self.data.text = text
        self.updateContentMarker()

    def getText(self):
        return(self.data.text)

    def setTitle(self, title):
        self.data.title = title

    def setEditWindow(self, editwindow):
        self.editwin = editwindow

    def setHeadlineColour(self, colourindex):
        #print ("Notewidget %i setHeadlineColour, index=%i" % (self.data.ident, colourindex))
        self.headlinecolour = self.scene.labelcolours[colourindex]
        self.headline.setBrush(Qt.QBrush(self.headlinecolour[0]))
        self.headlinetext.setDefaultTextColor(self.headlinecolour[1])
        self.data.setHeadlineColour(colourindex)

    def addLink(self, linkwidget):
        assert(isinstance(linkwidget, LinkWidget))
        self.linkwidgets.append(linkwidget)
        #self.data.addLink(link.data)
        linkwidget.adjust()

    def addNote(self, notewidget):
        self.subnotes.append(notewidget)
        self.contentmarker.setBrush(brush_yellow)
        self.updateContentMarker()

    def adjustLinks(self):
        for l in self.linkwidgets:
            l.adjust()

    def startOpenLink(self, pos):
        # if already creating a link terminate it first
        if (isinstance(self.scene.openlink, LinkWidget) and
            (self.scene.openlink != self)  # ignore self-referencing links
            and (self.scene.mode == 'NewLink')):
            self.scene.openlink.connect(self.scene.openlink.src, self)
        # start creating a Link
        self.scene.setMode('NewLink')
        # create new link, make a link-widget from it
        l = self.scene.data.newLink(self.data)
        lw = LinkWidget(l, self)
        # make it available upstream for mouse-move event
        self.scene.openlink = lw
        lw.adjust(pos)
        self.scene.addItem(lw)

    def closeOpenLink(self):
        ol = self.scene.openlink
        # ignore self-referencing links
        if isinstance(ol, LinkWidget):
            if (ol.src == self):
                # remove instance of Link from self.scene.data
                # remove ol from self.scene
                self.scene.removeItem(ol)
                self.scene.data.deleteLink(ol.data)
            else:
                ol.connect(ol.src, self)
                ol.titletext.setEdit(True)
                self.scene.p.signalDataChange()
        self.scene.openlink = None
        self.scene.setMode('Edit')

    def dragEnterEvent(self, e):
        print ('XXXXXXXXXXX', e)

    def dropEvent(self, e):
        print (self.data.title)
        print ('XXXXXXXXXXX', e)

    def mousePressEvent(self, e):
        assert(isinstance(e, QtWidgets.QGraphicsSceneMouseEvent))
        if e.button() == QtCore.Qt.RightButton: # RMB
            if self.headline.contains(e.pos()):
                e.ignore()
            else:
                self.startOpenLink(e.scenePos())
        elif (e.button() == QtCore.Qt.LeftButton):
            # print ('left-mousbutton-press', self.scene.mode)
            if (self.scene.mode == 'NewLink'):
                # i am the target of inserting a link, connect open link to me
                self.closeOpenLink()
            elif (self.scene.mode == 'Edit'):
                # ctrl-LMB = headline-edit
                if (e.modifiers() == QtCore.Qt.ControlModifier):
                    if self.headline.contains(e.pos()):
                        self.headlinetext.setEdit(True)
            else:
                # something is wrong
                print ("Warning: Scene in unknown mode, ignoring mouseclick")
        QtWidgets.QGraphicsRectItem.mousePressEvent(self, e)

    def mouseReleaseEvent(self, e):
        self.adjustLinks()
        self.scene.p.signalDataChange()
        #p = self.scenePos()
        #self.data.setGeometry(p.x(), p.y())  # update data with own position
        QtWidgets.QGraphicsRectItem.mouseReleaseEvent(self, e)

    def mouseMoveEvent(self, e):
        self.adjustLinks()
        # update links in all selected Notes for case of multi-move
        for nw in self.scene.selectedNotes():
            assert(isinstance(nw, NoteWidget))
            nw.adjustLinks()
        QtWidgets.QGraphicsRectItem.mouseMoveEvent(self, e)

    def mouseDoubleClickEvent(self, e):
        if self.headline.contains(e.pos()):
            self.scene.p.openTextEdit(self)
        else:
            self.scene.zoomIntoNote(self)

    def paint(self, painter, option, widget):
        super(NoteWidget,self).paint(painter, option, widget)
        self.headline.paint(painter, option, widget)

    def __str__(self):
        result = "NoteWidget:\n"
        result += "  title: %s\n" % self.data.title
        result += "  coords: %s\n" % str(self.rect().getCoords())
        result += "  %i links"
        return(result)

if __name__=='__main__':
    unittest.main()
